from typing import Any, Iterable, Dict, Union, Sequence
import os
import re
from sqlitedict import SqliteDict
from hilcat.core import RegistrableCache

def get_from_dict_or_default(d: dict, keys: list, default=None):
    for k in keys:
        if k in d:
            return d[k]
    return default

class SqliteDictCache(RegistrableCache):
    """
    Use a sqlite database as backend.
    """

    @classmethod
    def from_uri(cls, uri: str, **kwargs) -> 'SqliteDictCache':
        assert re.match(r'\w+:///.+', uri), uri
        schema, database = uri.split(':///')
        assert schema.lower() == 'sqlite', schema
        return cls(db_file=database, **kwargs)

    def __init__(
        self, db_file: Union[str, os.PathLike], autocommit: bool = True,
        scopes: Union[str, Dict[str, Union[str, Dict[str, Any]]], Sequence[Union[str, Dict[str, Any]]]] = None,
    ):
        os.makedirs(os.path.dirname(db_file), exist_ok=True)    # ensure db_file directory exists
        self._db_file = db_file
        self.autocommit = autocommit
        self._cache = {}
        if scopes is None:
            pass
        elif isinstance(scopes, str):
            self._init_scope(scopes)
        elif isinstance(scopes, dict):
            for scope, config in scopes.items():
                if isinstance(config, str):
                    self._init_scope(scope, table_name=config)
                elif isinstance(config, dict):
                    self._init_scope(scope, **config)
                else:
                    raise ValueError(f"Invalid scope config type: {type(config)}")
        else:
            for config in scopes:
                if isinstance(config, str):
                    self._init_scope(config)
                elif isinstance(config, dict):
                    self._init_scope(**config)
                else:
                    raise ValueError(f"Invalid scope config type: {type(config)}")

    def _init_scope(self, scope: str, **kwargs):
        table_name = get_from_dict_or_default(kwargs, ['table_name', 'tablename', 'table'], default=scope)
        autocommit = get_from_dict_or_default(kwargs, ['autocommit', 'auto_commit'], default=self.autocommit)
        self._cache[scope] = SqliteDict(self._db_file, tablename=table_name, autocommit=autocommit)

    def _has_scope_cache(self, scope: str) -> bool:
        return scope in self._cache

    def _get_scope_cache(self, scope: str) -> SqliteDict:
        if scope not in self._cache:
            self._init_scope(scope)
        return self._cache[scope]

    def exists(self, key: Any, scope: str = None, **kwargs) -> bool:
        if self._has_scope_cache(scope):
            d = self._get_scope_cache(scope)
            return key in d
        return False

    def fetch(self, key: Any, default: Any = None, scope: str = None, **kwargs) -> Any:
        if self._has_scope_cache(scope):
            d = self._get_scope_cache(scope)
            if key in d:
                return d[key]
        return default

    def set(self, key: Any, value: Any, scope: str = None, **kwargs) -> Any:
        d = self._get_scope_cache(scope)
        d[key] = value

    def pop(self, key: Any, scope: str = None, **kwargs) -> Any:
        if self._has_scope_cache(scope):
            d = self._get_scope_cache(scope)
            return d.pop(key)
        return None

    def scopes(self) -> Iterable[str]:
        return self._cache.keys()

    def keys(self, scope: str = None) -> Iterable[Any]:
        if self._has_scope_cache(scope):
            d = self._get_scope_cache(scope)
            return d.keys()
        return []




# -*- coding: utf-8 -*-

from typing import (
    Any, Iterable,
    Dict,
)
from ..core import RegistrableCache
import elasticsearch as es

class ElasticSearchCache(RegistrableCache):
    """
    Use elasticsearch as backend.
    Each scope corresponds to an index, and each key corresponds to a document.
    """

    @classmethod
    def from_uri(cls, uri: str, **kwargs) -> 'ElasticSearchCache':
        if not uri.startswith('es://'):
            raise ValueError(f"It's not a es uri: {uri}.")
        connect_args = kwargs.copy()
        connect_args['hosts'] = uri[5:]
        return cls(connect_args=connect_args)

    def __init__(self, client: es.Elasticsearch = None, connect_args: Dict[str, Any] = None):
        """
        Create a cache based on elasticsearch.
        :param client:              es client
        :param connect_args:        args to create a client
        """
        self.client = client or es.Elasticsearch(**(connect_args or {}))

    def exists(self, key: str, scope: str = None, **kwargs) -> bool:
        return self.client.exists(index=scope, id=key).body

    def fetch(self, key: str, default: Any = None, scope: str = None, **kwargs) -> Any:
        try:
            res = self.client.get(index=scope, id=key)
        except es.NotFoundError:
            return None
        return res.body.get('_source', default)

    def set(self, key: str, value: Dict[str, Any], scope: str = None, **kwargs) -> bool:
        res = self.client.index(index=scope, id=key, document=value)
        return res.body.get('result') in ['created', 'updated']

    def pop(self, key: str, scope: str = None, **kwargs) -> bool:
        res = self.client.delete(index=scope, id=key)
        return res.body.get('result') == 'deleted'

    def scopes(self) -> Iterable[str]:
        # it's not suggested to get all indexes in es
        raise NotImplementedError()

    def keys(self, scope: str = None) -> Iterable[str]:
        # it's not suggested to get all documents
        raise NotImplementedError()





# -*- coding: utf-8 -*-

try:
    from ._dist_ver import VERSION, __version__
except ImportError:
    from importlib_metadata import version, PackageNotFoundError
    try:
        __version__ = version('hilcat')
    except PackageNotFoundError:
        # package is not installed
        __version__ = "UNKNOWN"
    VERSION = __version__.split('.')



from abc import ABC, abstractmethod
from typing import Dict, Any

class ValueEncoder(ABC):
    """
    Encode cache value as dict with key to be column names.
    """
    @abstractmethod
    def encode(self, value: Any) -> Dict[str, Any]:
        pass

    def __call__(self, value: Any) -> Dict[str, Any]:
        return self.encode(value)

class NoModifyEncoder(ValueEncoder):
    def encode(self, value: Any) -> Dict[str, Any]:
        return value




# -*- coding: utf-8 -*-

from .version import __version__
from .core import (
    Cache, NoOpCache, MemoryCache,
    LocalFileCache, SimpleLocalFileCache,
    BinaryFileCache, SimpleBinaryFileCache,
    TextFileCache, SimpleTextFileCache,
    SimpleJsonFileCache,
    MiddleCache, MemoryMiddleCache,
    CacheAgent,
    register_backend,
)

try:
    from .db.redis import RedisCache
except ImportError:
    pass
else:
    register_backend('redis', RedisCache)

try:
    from .db.es import ElasticSearchCache
except ImportError:
    pass
else:
    register_backend('es', ElasticSearchCache)

from .db.relational import (
    SqlBuilder,
    SimpleSqlBuilder,
    QmarkSqlBuilder,
    NumericSqlBuilder,
    NamedSqlBuilder,
    FormatSqlBuilder,
    PyformatSqlBuilder,
    RelationalDbScopeConfig,
    RelationalDbCache,
    SingleTableConfig,
    SingleTableCache,
)

try:
    from .db.sqlite import (
        SqliteSqlBuilder,
        SqliteScopeConfig,
        SqliteCache,
        SqliteSingleTableCache,
    )
except ImportError:
    pass
else:
    register_backend('sqlite', SqliteCache)

try:
    from .db.postgresql import (
        PostgresqlBuilder,
        PostgresqlScopeConfig,
        PostgresqlCache,
        PostgresqlSingleTableCache,
    )
except ImportError:
    # psycopg maybe not installed
    pass
else:
    register_backend('postgresql', PostgresqlCache)

try:
    from .db.mysql import (
        MysqlSqlBuilder,
        MysqlScopeConfig,
        MysqlCache,
        MysqlSingleTableCache,
    )
except ImportError:
    # pymysql or mysql.connector should be installed
    pass
else:
    register_backend('mysql', MysqlCache)

try:
    # sqlitedict should be installed
    from .third.sqlitedict import SqliteDictCache
except ImportError:
    pass



from abc import ABC, abstractmethod
from typing import Dict, Any

class ValueDecoder(ABC):
    """
    Decode row selected from database.
    """

    @abstractmethod
    def encode(self, value: Dict[str, Any]) -> Any:
        pass

    def __call__(self, value: Dict[str, Any]) -> Any:
        return self.encode(value)

class NoModifyDecoder(ValueDecoder):
    def encode(self, value: Dict[str, Any]) -> Any:
        return value




# -*- coding: utf-8 -*-

import os
import re
import abc
import warnings
from typing import (
    List, Sequence, Any,
    Hashable, Dict, Union,
    Literal,
)
from .relational import (
    BaseRelationalDbCache,
    RelationalDbScopeConfig,
    RelationalDbCache,
    SingleTableCache,
    Operation,
    QmarkSqlBuilder, ValueAdapter,
)
import sqlite3

class SqliteSqlBuilder(QmarkSqlBuilder):
    default_column_type = "str"

    def build_select_all_table_operation(self) -> Operation:
        return Operation(statement="SELECT name FROM sqlite_master WHERE type = 'table'")

    def build_select_table_columns_operation(self, table: str, filter_uniq=False) -> Operation:
        return Operation(statement=f"PRAGMA table_info({table})")

    def get_column_name_from_result(self, result: Sequence[Any]) -> str:
        # second is column name
        return result[1]

class SqliteScopeConfig(RelationalDbScopeConfig):
    def __init__(self, scope: Hashable, table: str = None, uniq_column: str = None,
                 uniq_columns: Sequence[str] = ('id',), columns: Sequence[str] = ('data',),
                 column_types: Dict[str, str] = None,
                 value_adapter: Union[Literal['auto', 'default', 'single', 'tuple', 'list'], ValueAdapter] = 'auto',
                 default_column_type="str"):
        warnings.warn("use RelationalDbScopeConfig instead", DeprecationWarning)
        super().__init__(scope, table, uniq_column, uniq_columns, columns, column_types, value_adapter,
                         default_column_type)

class BaseSqliteCache(BaseRelationalDbCache, abc.ABC):
    """
    Base abstract class for sqlite backend.
    """

    api_module = sqlite3
    paramstyle = 'qmark'
    sql_builder = SqliteSqlBuilder()

    @classmethod
    def from_uri(cls, uri: str, **kwargs) -> 'BaseSqliteCache':
        assert re.match(r'\w+:///.+', uri), uri
        schema, database = uri.split(':///')
        return cls(database=database, **kwargs)

    def __init__(self, connection=None, database: str = None, connect_args: Dict[str, Any] = None):
        if database:
            # ensure database directory exists
            os.makedirs(os.path.dirname(database), exist_ok=True)
        super().__init__(connection, database, connect_args)

    def _get_unique_columns(self, table: str) -> List[Sequence[Any]]:
        columns = self._get_table_columns(table)
        return [x for x in columns if x[5] > 0]

    def _execute_many0(self, cursor, operation: Operation):
        # if executemany, no parameters should be given
        if operation.parameters:
            raise ValueError("there should be no parameters when invoke executemany()")
        # if semicolon in a stmt, it should be quoted
        stmts = operation.statement.split(";")
        for stmt in stmts:
            cursor.execute(stmt.strip())

class SqliteCache(BaseSqliteCache, RelationalDbCache):
    """
    Use a sqlite database as backend.
    """

class SqliteSingleTableCache(BaseSqliteCache, SingleTableCache):
    """
    Use one table of sqlite as backend.
    """



# -*- coding: utf-8 -*-

import abc
import warnings
from typing import (
    Any, Hashable, Sequence,
    Dict, Union, Literal,
)
from .relational import (
    Operation,
    FormatSqlBuilder,
    BaseRelationalDbCache,
    RelationalDbCache,
    SingleTableCache,
    RelationalDbScopeConfig,
    ValueAdapter,
)
import psycopg

class PostgresqlBuilder(FormatSqlBuilder):

    default_column_type = "text"

    def get_value_type(self, value: Any) -> str:
        return 's'

    def build_select_all_table_operation(self) -> Operation:
        return Operation(statement="SELECT tablename FROM pg_tables WHERE schemaname = 'public'")

    def build_select_table_columns_operation(self, table: str, filter_uniq=False) -> Operation:
        if filter_uniq:
            stmt = (f"SELECT a.attname FROM pg_index i"
                    f" JOIN pg_attribute a"
                    f" ON a.attrelid = i.indrelid AND a.attnum = any(i.indkey)"
                    f" WHERE i.indrelid = '{table}'::regclass AND i.indisprimary")
        else:
            stmt = f"SELECT column_name FROM information_schema.columns WHERE table_name = '{table}'"
        return Operation(statement=stmt)

class PostgresqlScopeConfig(RelationalDbScopeConfig):
    def __init__(self, scope: Hashable, table: str = None, uniq_column: str = None,
                 uniq_columns: Sequence[str] = ('id',), columns: Sequence[str] = ('data',),
                 column_types: Dict[str, str] = None,
                 value_adapter: Union[Literal['auto', 'default', 'single', 'tuple', 'list'], ValueAdapter] = 'auto',
                 default_column_type: str = None):
        warnings.warn("use RelationalDbScopeConfig instead", DeprecationWarning)
        super().__init__(scope, table, uniq_column, uniq_columns, columns, column_types, value_adapter,
                         default_column_type)

class BasePostgresqlCache(BaseRelationalDbCache, abc.ABC):
    """
    Base abstract class for postgresql backend.
    """

    api_module = psycopg
    paramstyle = "format"
    sql_builder = PostgresqlBuilder()

    def _execute_many0(self, cursor, operation: Operation):
        # Nothing happened when use cursor.executemany() to create multi tables.
        # So it's needed to split statement.

        # if executemany, no parameters should be given
        if operation.parameters:
            raise ValueError("there should be no parameters when invoke executemany()")
        # if semicolon in a stmt, it should be quoted
        stmts = operation.statement.split(";")
        for stmt in stmts:
            cursor.execute(stmt.strip())

class PostgresqlCache(BasePostgresqlCache, RelationalDbCache):
    """
    Use a postgresql database as backend.
    """

class PostgresqlSingleTableCache(BasePostgresqlCache, SingleTableCache):
    """
    Use one table of postgresql as backend.
    """



from abc import ABC, abstractmethod
from typing import Any, Dict, Sequence, Type, Union
import json

# cache data format may diff from db api, we need an adapter to bridging
class ValueAdapter(ABC):
    """
    Method `build_column_values` and `parse_column_values` should be reversed one-to-one mapping.
    That is to say:
        parse_column_values(build_column_values(x)) == x
        build_column_values(parse_column_values(x)) == x
    """
    @abstractmethod
    def build_column_values(self, value: Any) -> Dict[str, Any]:
        """
        Used in method `RelationalDbCache.set()` to build column values
        """
    @abstractmethod
    def parse_column_values(self, value: Dict[str, Any]) -> Any:
        """
        Used in method `RelationalDbCache.fetch()` to parse column values.
        """

class NoModifyAdapter(ValueAdapter):
    """
    Cache value should be exactly column values and thus nothing should do.
    """

    def build_column_values(self, value: Dict[str, Any]) -> Dict[str, Any]:
        return value

    def parse_column_values(self, value: Dict[str, Any]) -> Dict[str, Any]:
        return value

class SingleAdapter(ValueAdapter):
    """
    Cache value is exactly one column.
    """

    def __init__(self, col: str):
        self.col = col

    def build_column_values(self, value: Any) -> Dict[str, Any]:
        return {self.col: value}

    def parse_column_values(self, value: Dict[str, Any]) -> Any:
        return value[self.col]

class SingleJsonValueAdapter(SingleAdapter):
    """
    Cache value is stored in one column, and can be a complex type
      such as dict or list, type in db is text in json format.
    """
    def __init__(self, col: str, ensure_ascii=True):
        super().__init__(col)
        self.ensure_ascii = ensure_ascii

    def build_column_values(self, value: Dict[str, Any]) -> Dict[str, Any]:
        value = json.dumps(value, ensure_ascii=self.ensure_ascii)
        return super().build_column_values(value)

    def parse_column_values(self, value: Dict[str, Any]) -> Any:
        s = super().parse_column_values(value)
        return json.loads(s)

class JsonValueAdapter(ValueAdapter):
    """
    All value type in db should be text.
    """
    def __init__(self, ensure_ascii=True):
        super().__init__()
        self.ensure_ascii = ensure_ascii

    def build_column_values(self, value: Dict[str, Any]) -> Dict[str, Any]:
        return {k: json.dumps(v, ensure_ascii=self.ensure_ascii) for k, v in value.items()}

    def parse_column_values(self, value: Dict[str, Any]) -> Dict[str, Any]:
        return {k: json.loads(v) for k, v in value.items()}

class SequenceAdapter(ValueAdapter):
    """
    Cache value is a list or tuple, corresponding to some columns.
    """

    def __init__(self, cols: Sequence[str], return_type: Type[Union[list, tuple]] = tuple):
        self.cols = cols
        self.return_type = return_type

    def build_column_values(self, value: Sequence[Any]) -> Dict[str, Any]:
        return dict(zip(self.cols, value))

    def parse_column_values(self, value: Dict[str, Any]) -> Sequence[Any]:
        return self.return_type(map(value.get, self.cols))

class AutoAdapter(ValueAdapter):
    def __init__(self, cols: Sequence[str]):
        self.cols = list(cols)
        self.adapter = None

    def build_column_values(self, value: Any) -> Dict[str, Any]:
        # TODO 8/29/25
        # determine the actual adapter when first time see the data
        pass

    def parse_column_values(self, value: Dict[str, Any]) -> Any:
        # TODO 8/29/25
        pass

_BUILTIN_ADAPTER_BUILDERS = {
    'default': lambda cols: SingleAdapter(cols[0]) if len(cols) == 1 else NoModifyAdapter(),
    'immutable': lambda cols: NoModifyAdapter(),
    'single': lambda cols: SingleAdapter(cols[0]),
    'tuple': lambda cols: SequenceAdapter(cols, return_type=tuple),
    'list': lambda cols: SequenceAdapter(cols, return_type=list),
    'json': lambda cols: SingleJsonValueAdapter(cols[0]) if len(cols) == 1 else JsonValueAdapter(),
}




# -*- coding: utf-8 -*-

"""
Core code for the high level abstract cache.
"""

from typing import (
    Any, Dict, List,
    Callable, Iterable,
    Hashable, Optional,
    Type, Union, Sequence,
)
from abc import (
    ABC, abstractmethod,
)
import os
import sys
import json
import inspect
import pathlib
import warnings
import builtins
import functools
import urllib.parse

def _create_fn(name: str, args: List[str], body: List[str], *,
               _globals: Dict[str, Any] = None,
               _locals: Dict[str, Any] = None,
               return_type: Optional[Type] = None):
    _globals = _globals or {}
    # Note that we mutate locals when exec() is called.  Caller
    # beware!  The only callers are internal to this module, so no
    # worries about external callers.
    if _locals is None:
        _locals = {}
    if 'BUILTINS' not in _locals:
        _locals['BUILTINS'] = builtins
    return_annotation = ''
    if return_type is not None:
        _locals['_return_type'] = return_type
        return_annotation = '->_return_type'
    args = ','.join(args)
    body = '\n'.join(f'  {b}' for b in body)

    # Compute the text of the entire function.
    txt = f' def {name}({args}){return_annotation}:\n{body}'

    local_vars = ', '.join(_locals.keys())
    txt = f"def __create_fn__({local_vars}):\n{txt}\n return {name}"

    ns = {}
    exec(txt, _globals, ns)
    return ns['__create_fn__'](**_locals)

class Storage(ABC):
    """
    Base storage class, all api is equal to the cache except method load() and save().

    Any key is under a scope, but scope can be always same value or None in some implements.

    The key is recommended to use an uniq id as key, but actually it can be anything.
    The scope is recommended to use a string.
    """

    def close(self):
        """
        For some remote server, close the connection.

        For some local file cache, close and save files.
        """

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def __getitem__(self, item):
        # just fetch value from default scope
        return self.fetch(item)

    def __setitem__(self, key, value):
        # just set value as to default scope
        self.set(key, value)

    @abstractmethod
    def exists(self, key: Any, scope: Any = None, **kwargs) -> bool:
        """
        Check if given key exists in the cache.
        """

    @abstractmethod
    def fetch(self, key: Any, default: Any = None, scope: Any = None, **kwargs) -> Any:
        """
        Fetch value of the given key.
        Maybe raise an error or return a default value when key not exists.
        """

    @abstractmethod
    def set(self, key: Any, value: Any, scope: Any = None, **kwargs) -> Any:
        """
        Set value for the given key.
        :return:    Can be old value, merged value, a bool or anything depends on the implement.
        """

    def update(self, key: Any, value: Any, scope: Any = None, **kwargs) -> Any:
        """
        Update value for the given key.
        """
        # default is same as set
        return self.set(key, value, scope=scope, **kwargs)

    def get(self, key: Any, func: Callable = None, func_args: Sequence[Any] = None,
            func_kwargs: Dict[str, Any] = None, scope: Any = None, **kwargs) -> Any:
        """
        If func is None, it's equal to method `fetch()`.
        If func not None,
            get value from cache if key exists;
            else run the func and save value to the cache.
        """
        if func is None or self.exists(key, scope=scope):
            return self.fetch(key, scope=scope)
        func_args = func_args or []
        func_kwargs = func_kwargs or {}
        value = func(*func_args, **func_kwargs)
        self.set(key, value, scope=scope)
        return value

    @abstractmethod
    def pop(self, key: Any, scope: Any = None, **kwargs) -> Any:
        """
        Delete value for the given key.
        :return:    Value in the cache if arg return_value is `True`.
        """

    def scopes(self) -> Iterable[Any]:
        """
        Get all scopes in the cache.
        """
        raise NotImplementedError("It's not allowed to get all scopes.")

    def keys(self, scope: Any = None) -> Iterable[Any]:
        """
        Get all keys in the cache.
        """
        raise NotImplementedError("It's not allowed to get all keys.")

    def __call__(self, f=None, scope=None, ignore_first_arg=False, make_key_func=None, **kwargs):
        """
        Decorate function with a cache.
        To cache result of the function.
        :param f:           function to be decorated
        :param scope:       scope for cache to get and set
        :param ignore_first_arg:    whether to ignore function first arg when gen cache key;
                                    if function is a class method, this can be set as `True`
        :param make_key_func:   a function to gen key from arg values
        """

        def wrap(_f):
            def get_var(var1: str, var2: str):
                return var1 if var1 != _f.__name__ else var2

            sig = inspect.signature(_f)
            parameters = list(sig.parameters.keys())
            if ignore_first_arg:
                parameters = parameters[1:]
            if len(parameters) == 0:
                raise ValueError("Function consume no arg.")

            has_kwargs = list(sig.parameters.values())[-1].kind.value == 5
            make_key_var = get_var("_make_key", "__make_key")
            make_key = make_key_func
            if not make_key:
                if len(parameters) == 1:
                    if has_kwargs:
                        def make_key(*args, **kwargs):
                            bind = sig.bind(*args, **kwargs)
                            return tuple(sorted(bind.arguments[parameters[0]].items()))
                    else:
                        def make_key(*args, **kwargs):
                            bind = sig.bind(*args, **kwargs)
                            return bind.arguments[parameters[0]]
                else:
                    if has_kwargs:
                        def make_key(*args, **kwargs):
                            bind = sig.bind(*args, **kwargs)
                            return (tuple((x, bind.arguments[x]) for x in parameters[:-1]) +
                                    tuple(sorted(bind.arguments[parameters[-1]].items())))
                    else:
                        def make_key(*args, **kwargs):
                            bind = sig.bind(*args, **kwargs)
                            return tuple(map(bind.arguments.get, parameters))

            _globals = sys.modules[_f.__module__].__dict__
            func_var = get_var("_func", "__func")
            score_var = get_var("_scope", "_scope")
            cache_var = get_var("_cache", "__cache")
            _locals = {
                "BUILTINS": builtins,
                func_var: _f,
                score_var: scope,
                cache_var: self,
                make_key_var: make_key,
            }
            args = ["*args", "**kwargs"]
            body_lines = [
                f"return {cache_var}.get({make_key_var}(*args, **kwargs),"
                f" lambda: {func_var}(*args, **kwargs), scope={score_var})"
            ]
            return_type = _f.__annotations__.get('return')
            func = _create_fn(_f.__name__, args, body_lines,
                              _globals=_f.__globals__,
                              _locals=_locals,
                              return_type=return_type)
            func = functools.update_wrapper(func, _f)
            return func

        # See if we're being called as @Cache() or @Cache()().
        if f is None:
            # We're called with 2 pair of parens.
            return wrap

        # We're called with 1 pair of parens.
        return wrap(f)

class Cache(Storage, ABC):
    """
    An abstract cache interface.
    """

    @classmethod
    def from_uri(cls, uri: str, **kwargs) -> 'Cache':
        """
        Create a cache base on given backend.
        Auto-detect backend engine.
        Subclass should override this method.
        """
        r = urllib.parse.urlsplit(uri)
        backend = BACKENDS.get(r.scheme, DEFAULT_BACKEND)
        if backend is None:
            if r.scheme:
                raise ValueError(f"schema not given: {uri}")
            raise ValueError(f"Unsupported backend: {r.scheme}")
        if isinstance(backend, Cache):
            # engine is a cache instance, return it directly
            return backend
        if inspect.isclass(backend) and issubclass(backend, Cache):
            # engine is a cache subclass, create it
            return backend.from_uri(uri, **kwargs)
        if callable(backend):
            # engine is a function, call it
            # should inspect function signature and pass different args ?
            try:
                return backend(uri, **kwargs)
            except TypeError:
                return backend(uri, kwargs)
        raise RuntimeError(f"Unexpected engine type: {type(backend)}")

    def load(self, scopes: Iterable[Any] = None, **kwargs):
        """
        Load scope data from persistence storage.
        """

    def backup(self, scopes: Iterable[Any] = None, **kwargs):
        """
        Save scope data to persistence storage.

        If cache is just a temp storage, this method should be implemented.
        """

    def save(self, scopes: Iterable[Any] = None, **kwargs):
        # backup has no ambiguity, save is going to be deprecated
        warnings.warn("use backup() instead", DeprecationWarning)
        self.backup(scopes, **kwargs)

# schema -> cache builder; to build a cache from uri
_BACKEND_TYPE = Union[Cache, Type[Cache], Callable[[str, Dict[str, Any]], Cache]]
BACKENDS: Dict[str, _BACKEND_TYPE] = {}
DEFAULT_BACKEND: Optional[_BACKEND_TYPE] = None

def register_backend(schema: str, backend: _BACKEND_TYPE):
    """
    Register a backend, then we can create the cache by `Cache.from_uri()`.
    """
    if schema in BACKENDS:
        warnings.warn(f"Backend {schema} already defined, it will be overwritten.")
    BACKENDS[schema] = backend

class RegistrableCache(Cache, ABC):
    """
    Subclasses are forced to implement the `from_uri` method.
    """
    @classmethod
    @abstractmethod
    def from_uri(cls, uri: str, **kwargs) -> Cache:
        pass

class NoOpCache(Cache):
    """
    A cache placeholder, used when all values are calculated currently but may need a cache some day.
    """
    def exists(self, key: Any, scope: Any = None, **kwargs) -> bool:
        return False

    def fetch(self, key: Any, default: Any = None, scope: Any = None, **kwargs) -> Any:
        return None

    def set(self, key: Any, value: Any, scope: Any = None, **kwargs) -> None:
        pass

    def update(self, key: Any, value: Any, scope: Any = None, **kwargs) -> Any:
        return None

    def pop(self, key: Any, scope: Any = None, **kwargs) -> Any:
        return None

class MemoryCache(Cache):
    """
    Use a dict as backend.
    """
    def __init__(self) -> None:
        self.data = {}

    def exists(self, key: Hashable, scope: Hashable = None, **kwargs) -> bool:
        return scope in self.data and key in self.data[scope]

    def fetch(self, key: Hashable, default: Any = None, scope: Hashable = None, **kwargs) -> Any:
        return self.data.get(scope, {}).get(key, default)

    def set(self, key: Hashable, value: Any, scope: Hashable = None, **kwargs) -> bool:
        if scope not in self.data:
            self.data[scope] = {}
        self.data[scope][key] = value
        return True

    def update(self, key: Hashable, value: Any, scope: Hashable = None, **kwargs) -> Any:
        old = None
        if self.exists(key, scope=scope):
            old = self.fetch(key, scope=scope)
        self.set(key, value, scope=scope)
        return old

    def pop(self, key: Any, scope: Any = None, **kwargs) -> Any:
        if self.exists(key, scope=scope):
            return self.data[scope].pop(key)
        return None

    def scopes(self) -> Iterable[Hashable]:
        return self.data.keys()

    def keys(self, scope: Any = None) -> Iterable[Hashable]:
        return self.data.get(scope, {}).keys()

class LocalFileCache(Cache):
    """
    Each key corresponds to a file.
    """
    @abstractmethod
    def _get_filepath(self, key: Any, scope: Any = None) -> str:
        """
        Get filepath for the given key to save value.
        """

    def exists(self, key: Any, scope: Any = None, **kwargs) -> bool:
        filepath = self._get_filepath(key, scope=scope)
        return os.path.exists(filepath)

    @abstractmethod
    def _read_file(self, filepath: str) -> Any:
        """
        Read file content.
        """

    def fetch(self, key: Any, default: Any = None, scope: Any = None, **kwargs) -> Any:
        """
        If file exists, read file content; else, return default.
        """
        filepath = self._get_filepath(key, scope=scope)
        if os.path.exists(filepath):
            return self._read_file(filepath)
        return default

    def _write_file0(self, filepath: str, content: Any):
        """
        The actual write file method.
        """
        # Someone may overwrite _write_file() and this method would not be called.
        # If it's not necessary, there is no impact for raising an error.
        raise NotImplementedError()

    def _write_file(self, filepath: str, content: Any):
        """
        Write content to file.
        """
        # create parent dir first
        os.makedirs(os.path.abspath(os.path.dirname(filepath)), exist_ok=True)
        self._write_file0(filepath, content)

    def set(self, key: Any, value: Any, scope: Any = None, **kwargs) -> bool:
        filepath = self._get_filepath(key, scope=scope)
        self._write_file(filepath, value)
        return True

    def update(self, key: Any, value: Any, scope: Any = None, return_old=False, **kwargs) -> Any:
        filepath = self._get_filepath(key, scope=scope)
        result = None
        if return_old and os.path.exists(filepath):
            result = self._read_file(filepath)
        self._write_file(filepath, value)
        return result

    def pop(self, key: Any, scope: Any = None, return_old=False, **kwargs) -> Any:
        filepath = self._get_filepath(key, scope=scope)
        result = None
        if os.path.exists(filepath):
            if return_old:
                result = self._read_file(filepath)
            os.remove(filepath)
        return result

class BinaryFileCache(LocalFileCache, ABC):
    """
    Each value stored in a binary file.
    """
    def _read_file(self, filepath: str) -> Any:
        with open(filepath, 'rb') as f:
            return f.read()

    def _write_file0(self, filepath: str, content: Any):
        with open(filepath, 'wb') as f:
            f.write(content)

class TextFileCache(LocalFileCache, ABC):
    """
    Each value stored in a text file.
    """
    def __init__(self, encoding='utf-8'):
        """
        :param encoding:        file encoding
        """
        self.encoding = encoding

    def _read_file(self, filepath: str) -> str:
        with open(filepath, encoding=self.encoding) as f:
            return f.read()

    def _write_file0(self, filepath: str, content: str):
        with open(filepath, 'w', encoding=self.encoding) as f:
            f.write(content)

class SimpleLocalFileCache(LocalFileCache, ABC):
    """
    All files saved in a directory, and key to be relative path without suffix.
    """
    def __init__(self, root_dir: str, suf: str = ''):
        """
        :param root_dir:    root dir for files
        :param suf:         commonly should start with '.'
        """
        self.root_dir = root_dir
        self.suf = suf

    def _get_filepath(self, key: str, scope: str = None) -> str:
        return os.path.join(self.root_dir, scope or '', key + self.suf)

    def keys(self, scope: str = None) -> Iterable[str]:
        # find all files under the directory
        root = pathlib.Path(self.root_dir)
        if scope:
            root = root.joinpath(scope)
        for f in root.rglob("*" + self.suf):
            if f.is_file():
                key = f.relative_to(root).as_posix()
                if self.suf:
                    key = key[:-len(self.suf)]
                yield key

class SimpleBinaryFileCache(SimpleLocalFileCache, BinaryFileCache):
    def __init__(self, root_dir: str, suf: str = ''):
        SimpleLocalFileCache.__init__(self, root_dir, suf)
        BinaryFileCache.__init__(self)

class SimpleTextFileCache(SimpleLocalFileCache, TextFileCache):
    def __init__(self, root_dir: str, suf: str = '', encoding='utf-8'):
        SimpleLocalFileCache.__init__(self, root_dir, suf)
        TextFileCache.__init__(self, encoding)

class SimpleJsonFileCache(SimpleTextFileCache):
    def __init__(self, root_dir: str, suf: str = '.json', encoding='utf-8'):
        super().__init__(root_dir, suf, encoding)

    def _read_file(self, filepath: str) -> Union[dict, list]:
        with open(filepath, encoding=self.encoding) as f:
            return json.load(f)

    def _write_file0(self, filepath: str, content: Union[dict, list]):
        with open(filepath, 'w', encoding=self.encoding) as f:
            json.dump(content, f)

class MiddleCache(Cache, ABC):
    """
    A cache which has a persistent storage.

    The storage is also a cache but ignore scope.
    The key of the storage corresponds to the scope of this cache.
    """

    def __init__(self, storage: Storage):
        self.storage = storage

    @abstractmethod
    def set_scope_values(self, data: Any, scope=None):
        """
        Set values for given scope.
        """

    def load(self, scopes: Iterable[Any] = None, **kwargs):
        if scopes is None:
            # It mab be not allowed to get all keys for the storage.
            scopes = self.storage.keys()
        for scope in scopes:
            values = self.storage.fetch(scope)
            self.set_scope_values(values, scope=scope)

    @abstractmethod
    def get_scope_values(self, scope=None) -> Any:
        """
        Get values for given scope.
        """

    def backup(self, scopes: Iterable[Any] = None, **kwargs):
        if scopes is None:
            scopes = self.scopes()
        for scope in scopes:
            values = self.get_scope_values(scope)
            self.storage.set(scope, values)

class MemoryMiddleCache(MemoryCache, MiddleCache):
    """
    Data cached in memory with a persistent storage.
    """

    def __init__(self, storage: Storage):
        MemoryCache.__init__(self)
        MiddleCache.__init__(self, storage)

    def set_scope_values(self, data: Dict[Hashable, Any], scope: Hashable = None):
        self.data[scope] = data

    def get_scope_values(self, scope: Hashable = None) -> Dict[Hashable, Any]:
        return self.data.get(scope, {})


class CacheAgent(Cache):
    """
    Use multi cache as backends when write data.
    First cache as backend when read data.
    """

    def __init__(self, backend: Cache, *extra_writers: Cache):
        self.backend = backend
        self.extra_writers = extra_writers

    def exists(self, key: Any, scope: Any = None, **kwargs) -> bool:
        return self.backend.exists(key, scope=scope, **kwargs)

    def fetch(self, key: Any, default: Any = None, scope: Any = None, **kwargs) -> Any:
        return self.backend.fetch(key, default=default, scope=scope, **kwargs)

    def set(self, key: Any, value: Any, scope: Any = None, **kwargs) -> Any:
        for backend in self.extra_writers:
            backend.set(key, value, scope=scope, **kwargs)
        return self.backend.set(key, value, scope=scope, **kwargs)

    def update(self, key: Any, value: Any, scope: Any = None, **kwargs) -> Any:
        for backend in self.extra_writers:
            backend.update(key, value, scope=scope, **kwargs)
        return self.backend.update(key, value, scope=scope, **kwargs)

    def pop(self, key: Any, scope: Any = None, **kwargs) -> Any:
        for backend in self.extra_writers:
            backend.pop(key, scope=scope, **kwargs)
        return self.backend.pop(key, scope=scope, **kwargs)

    def scopes(self) -> Iterable[Any]:
        return self.backend.scopes()

    def keys(self, scope: Any = None) -> Iterable[Any]:
        return self.backend.keys(scope=scope)




# -*- coding: utf-8 -*-

from typing import Any, Iterable, Union
import redis
from ..core import RegistrableCache

_REDIS_KEY_TYPE = Union[str, bytes]

class RedisCache(RegistrableCache):
    """
    Use redis as backend.
    Ignore scope for all cache methods.
    """

    @classmethod
    def from_uri(cls, uri: str, **kwargs) -> 'RedisCache':
        return cls(url=uri, **kwargs)

    def __init__(self, client: redis.Redis = None, url: str = None, host: str = None, port: int = None, db=0):
        if client is not None:
            self.client = client
        elif url is not None:
            self.client = redis.from_url(url)
        elif host is not None:
            self.client = redis.Redis(host=host, port=port, db=db)
        else:
            raise ValueError("One of client, url or host should given.")

    def exists(self, key: _REDIS_KEY_TYPE, scope: Any = None, **kwargs) -> bool:
        return self.client.exists(key) > 0

    def fetch(self, key: _REDIS_KEY_TYPE, default: Any = None, scope: Any = None, **kwargs) -> Any:
        value = self.client.get(key)
        if value is None:
            return default
        return value

    def set(self, key: _REDIS_KEY_TYPE, value: Any, scope: Any = None, **kwargs) -> Any:
        self.client.set(key, value, **kwargs)

    def pop(self, key: _REDIS_KEY_TYPE, scope: Any = None, **kwargs) -> Any:
        # delete a key
        self.client.delete(key)

    def keys(self, scope: Any = None) -> Iterable[_REDIS_KEY_TYPE]:
        return self.client.keys()




from typing import (
    Any, Dict, List,  Sequence, Literal, Iterable,
    Union, Callable,
)
import re
from abc import ABC
from types import ModuleType
import warnings
from hilcat.core import RegistrableCache
from .config import Operation, BaseTableConfig, RelationalDbScopeConfig, SingleTableConfig
from .sql import SqlBuilder, DEFAULT_SQL_BUILDERS

_FETCH_SIZE_TYPE = Union[Literal['one', 'all'], int]
_COLUMN_VALUE_TYPE = Union[str, int]
_KEY_TYPE = Union[_COLUMN_VALUE_TYPE, Sequence[_COLUMN_VALUE_TYPE]]

class BaseRelationalDbCache(RegistrableCache, ABC):
    # if given api_module, use the module to connect
    api_module: ModuleType

    # parameter marker format described in pep-0249
    paramstyle: Literal['qmark', 'numeric', 'named', 'format', 'pyformat']

    # how to build fetch, update and delete sql
    # all sql required by cache should be generated from the builder
    sql_builder: SqlBuilder

    def __init_subclass__(cls):
        super().__init_subclass__()

        # if paramstyle not specified, use value in module
        if ((not hasattr(cls, 'paramstyle') or cls.paramstyle is None)
                and (hasattr(cls, 'api_module') and cls.api_module is not None)):
            cls.paramstyle = cls.api_module.paramstyle

        # if sql_builder not specified, use default
        if ((not hasattr(cls, 'sql_builder') or cls.sql_builder is None)
                and (hasattr(cls, 'paramstyle') and cls.paramstyle is not None)):
            if cls.paramstyle not in DEFAULT_SQL_BUILDERS:
                raise ValueError(f"Wrong paramstyle: {cls.paramstyle}")
            cls.sql_builder = DEFAULT_SQL_BUILDERS[cls.paramstyle]

    def connect_db(self, database: str = None, connect_args: Dict[str, Any] = None):
        """
        Connect to a database, return a connection object described in pep-0249.

        If always given a connection when init, this method may never run.
        """
        return self.api_module.connect(database, **(connect_args or {}))

    @classmethod
    def from_uri(cls, uri: str, **kwargs) -> 'BaseRelationalDbCache':
        assert re.match(r'\w+://.+', uri), uri
        schema, database = uri.split('://')
        return cls(database=database, **kwargs)

    def __init__(self, connection=None, database: str = None, connect_args: Dict[str, Any] = None):
        """
        Create a cache based on relational database.
        :param connection:          connection to the database
        :param database:            uri for the database
        :param connect_args:        custom connect args
        """
        self._database = database
        self._connect_args = connect_args
        self._conn = connection
        self._cursor = None

    @property
    def conn(self):
        if self._conn is None:
            self._conn = self.connect_db(self._database, self._connect_args)
        return self._conn

    @property
    def cursor(self):
        if self._cursor is None:
            self._cursor = self.conn.cursor()
        return self._cursor

    def close(self):
        if self._cursor is not None:
            self._cursor.close()
        if self._conn is not None:
            self._conn.close()

    def _create_table_if_not_exists(self, *tables: BaseTableConfig):
        operations = [self.sql_builder.build_create_table_operation(config, check_exists=True)
                      for config in tables]
        self._execute(*operations, cursor='new', commit=True)

    def _fetch_data(self, cursor, size: _FETCH_SIZE_TYPE = None) -> Any:
        if size is None:
            size = 0
        elif size == 'one':
            size = 1
        elif size == 'all':
            size = -1
        if not isinstance(size, int):
            raise ValueError(f"Unexpected size type: {type(size)}")
        if size == 0:
            return None
        elif size == 1:
            return cursor.fetchone()
        elif size < 0:
            return cursor.fetchall()
        else:
            return cursor.fetchmany(size)

    def _execute_many0(self, cursor, operation: Operation):
        """
        Some api not allow executemany(), in this case, overwrite this method.
        """
        cursor.executemany(operation.statement, operation.parameters)

    def _execute(self, *operations: Union[str, Operation], cursor='new', auto_close_cursor=True,
                 fetch_size: _FETCH_SIZE_TYPE = None, commit=False) -> Any:
        """
        Execute sql and fetch result.
        :param operations:      sequence of operation, same as described in pep-0249
        :param cursor:          if `None`, use global; if 'new', create a new cursor; else, use the given cursor
        :param auto_close_cursor:   close cursor in the end
        :param fetch_size:      how many rows should return
        :param commit:          do commit to database or not
        """
        close_cursor = False
        if cursor is None:
            cursor = self.cursor
        elif cursor == 'new':
            cursor = self.conn.cursor()
            close_cursor = auto_close_cursor
        for operation in operations:
            if isinstance(operation, str):
                operation = Operation(statement=operation)
            if operation.many:
                self._execute_many0(cursor, operation)
            else:
                cursor.execute(operation.statement, operation.parameters)
        result = self._fetch_data(cursor, size=fetch_size)
        if commit:
            self.conn.commit()
        if close_cursor:
            # close the cursor if it's created in this method
            cursor.close()
        return result

    def _get_all_table_names_in_db(self) -> List[str]:
        """
        Get all tables in the database, used when init scopes.
        """
        operation = self.sql_builder.build_select_all_table_operation()
        # assume first is table name
        return [x[0] for x in self._execute(operation, fetch_size='all')]

    def _get_table_columns(self, table: str) -> List[Sequence[Any]]:
        """
        Get columns for given table.
        :return:    column names
        """
        operation = self.sql_builder.build_select_table_columns_operation(table)
        columns = self._execute(operation, fetch_size='all')
        return list(columns)

    def _get_table_column_names(self, table: str) -> List[str]:
        """
        Get column names for given table.
        """
        columns = self._get_table_columns(table)
        return [self.sql_builder.get_column_name_from_result(x) for x in columns]

    def _get_unique_columns(self, table: str) -> List[Sequence[Any]]:
        """
        Get uniq columns for given table.
        """
        operation = self.sql_builder.build_select_table_columns_operation(table, filter_uniq=True)
        columns = self._execute(operation, fetch_size='all')
        return list(columns)

    def _get_unique_column_names(self, table: str) -> List[str]:
        """
        Get uniq column names for given table.
        """
        columns = self._get_unique_columns(table)
        return [self.sql_builder.get_column_name_from_result(x) for x in columns]

    def _get_unique_column_name(self, table: str) -> str:
        """
        Get uniq column name for given table.
        """
        warnings.warn("Deprecated, use _get_unique_column_names() instead", DeprecationWarning)
        columns = self._get_unique_column_names(table)
        if len(columns) != 1:
            raise ValueError(f"There should be exactly one uniq column, but {len(columns)} has given.")
        return columns[0]

    def _exists(self, key: Sequence[Any], table: BaseTableConfig) -> bool:
        """
        Actual method to test if a key exists in the table.
        :param key:         value of uniq columns
        :param table:       table config
        """
        operation = self.sql_builder.build_select_operation(
            config=table, key=key, limit=1,
            select_columns=table.uniq_columns,
        )
        return self._execute(operation, fetch_size=1) is not None

    def _fetch(self, key: Sequence[Any], table: BaseTableConfig, default: Any = None) -> Any:
        """
        Actual method to fetch data from table.
        :param key:         value of uniq columns
        :param table:       table config
        :param default:     if select no data, return this value
        """
        operation = self.sql_builder.build_select_operation(
            config=table, key=key, limit=1,
            select_columns=table.columns,
        )
        row = self._execute(operation, fetch_size=1)
        if row is None:
            return default
        value = dict(zip(table.columns, row))
        return table.decoder(value)

    def _set(self, key: Sequence[Any], value: Any, table: BaseTableConfig) -> bool:
        """
        Actual method to update or insert row into table.
        :param key:         value of uniq columns
        :param value:       value of all other columns
        :param table:       table config
        """
        value = table.encoder(value)
        row = {}
        for name, k in zip(table.uniq_columns, key):
            if name in value and k != value[name]:
                raise ValueError(f"column {name} key {k} is different from value {value[name]}")
            row[name] = k
        row.update((name, value[name]) for name in table.columns if name in value)
        operation = self.sql_builder.build_update_operation(config=table, key=key, value=row)
        self._execute(operation, cursor='new', auto_close_cursor=True, commit=True)
        return True

    def _pop(self, key: Sequence[Any], table: BaseTableConfig):
        """
        Actual method to remove row from table.
        :param key:         value of uniq columns
        :param table:       table config
        """
        operation = self.sql_builder.build_delete_operation(config=table, key=key)
        self._execute(operation, cursor='new', auto_close_cursor=True, commit=True)

_SCOPES_ARG_TYPE = Union[
    None,
    str,    # single scope cache
    RelationalDbScopeConfig,
    Dict[str, Union[
        str,        # specify table name
        Dict[str, Any],
        RelationalDbScopeConfig
    ]],
    Sequence[Union[
        str,    # only specify scope
        Dict[str, Any],
        RelationalDbScopeConfig
    ]],
]

class RelationalDbCache(BaseRelationalDbCache, ABC):
    """
    Use a relational database as backend.
    Each scope corresponds to a table, and each key corresponds to a row.
    It's recommended to use a string as key, but other type such as int is also allowed.
    """

    @classmethod
    def from_uri(cls, uri: str, **kwargs) -> 'RelationalDbCache':
        assert re.match(r'\w+://.+', uri), uri
        schema, database = uri.split('://')
        return cls(database=database, **kwargs)

    def __init__(
        self, connection=None, database: str = None, connect_args: Dict[str, Any] = None,
        scopes: _SCOPES_ARG_TYPE = None,
        new_scope_config: Callable[[str], RelationalDbScopeConfig] = None,
        all_table_as_scope=False
    ):
        """
        Create a cache based on relational database.
        :param connection:          connection to the database
        :param database:            uri for the database
        :param connect_args:        custom connect args
        :param scopes:              initialized scopes
        :param new_scope_config:    when a new scope given, how to config it
        :param all_table_as_scope:  if `True`, add all table in database to scopes
        """
        super().__init__(connection, database, connect_args)
        self._scopes = {}   # scope -> config
        self._tables = {}   # table -> config, should correspond to _scopes
        self._new_scope_config = new_scope_config
        self._init_scopes(scopes=self._process_scopes_arg(scopes), all_table_as_scope=all_table_as_scope)

    def _process_scopes_arg(self, scopes: _SCOPES_ARG_TYPE) -> List[RelationalDbScopeConfig]:
        """
        Process scopes arg given in __init__ function as a config list.
        """
        if scopes is None:
            return []
        elif isinstance(scopes, str):
            return [RelationalDbScopeConfig(scopes)]
        elif isinstance(scopes, RelationalDbScopeConfig):
            return [scopes]
        elif isinstance(scopes, dict):
            result = []
            for scope, config in scopes.items():
                if isinstance(config, str):
                    config = RelationalDbScopeConfig(scope, table=config)
                elif isinstance(config, dict):
                    if 'scope' in config and config['scope'] != scope:
                        raise ValueError(f"conflict scope: {scope} {config['scope']}")
                    config.pop('scope', None)
                    config = RelationalDbScopeConfig(scope, **config)
                if not isinstance(config, RelationalDbScopeConfig):
                    raise ValueError(f"Invalid scope config type: {type(config)}")
                if config.scope != scope:
                    raise ValueError(f"conflict scope: {scope} {config.scope}")
                result.append(config)
            return result
        else:
            result = []
            for config in scopes:
                if isinstance(config, str):
                    config = RelationalDbScopeConfig(config)
                elif isinstance(config, dict):
                    config = RelationalDbScopeConfig(**config)
                if not isinstance(config, RelationalDbScopeConfig):
                    raise ValueError(f"Invalid scope config type: {type(config)}")
                result.append(config)
            return result

    def _add_scope(self, config: RelationalDbScopeConfig):
        # scope and table should be uniq
        if config.scope in self._scopes:
            raise ValueError(f"duplicated scope: {config.scope}")
        if config.table in self._tables:
            raise ValueError(f"duplicated table: {config.table}")
        self._scopes[config.scope] = config
        self._tables[config.table] = config

    def _init_scopes(self, scopes: List[RelationalDbScopeConfig], all_table_as_scope: bool):
        # if some scopes have given, add to the scope mapper
        # if table not exists, create it
        if scopes:
            for config in scopes:
                self._add_scope(config)
            self._create_table_if_not_exists(*scopes)

        # find tables in database, if table not configured, add to the scope mapper
        if all_table_as_scope:
            # select all tables
            table_names = self._get_all_table_names_in_db()

            # table should not bound to a scope, and should not be same as a scope, remove these tables
            table_names = [x for x in table_names if x not in self._tables and x not in self._scopes]

            # for retain tables, add to the cache
            for table in table_names:
                columns = self._get_table_column_names(table)
                uniq_columns = self._get_unique_column_names(table)
                self._add_scope(RelationalDbScopeConfig(
                    scope=table,    # use table name as scope
                    table=table,
                    uniq_columns=uniq_columns,     # uniq column as id
                    columns=columns,
                ))

    def _get_scope_config(self, scope: str) -> RelationalDbScopeConfig:
        if scope in self._scopes:
            return self._scopes[scope]
        if not self._new_scope_config:
            raise ValueError(f"new scope is not allowed: {scope}")
        config = self._new_scope_config(scope)
        config = self._process_scopes_arg({scope: config})[0]
        if scope != config.scope:
            raise ValueError(f"conflict scope: {scope} {config.scope}")
        self._add_scope(config)
        self._create_table_if_not_exists(config)    # create table if not exists
        return config

    def _check_key(self, key: _KEY_TYPE):
        if key is None:
            raise ValueError("Arg key should not be None.")

    def exists(self, key: _KEY_TYPE, scope: str = None, **kwargs) -> bool:
        self._check_key(key)
        config = self._get_scope_config(scope)
        key = config.normalize_uniq_column_values(key)
        return self._exists(key, config)

    def fetch(self, key: _KEY_TYPE, default: Any = None, scope: Any = None, **kwargs) -> Any:
        self._check_key(key)
        config = self._get_scope_config(scope)
        key = config.normalize_uniq_column_values(key)
        return self._fetch(key, config, default=default)

    def set(self, key: _KEY_TYPE, value: Any, scope: Any = None, **kwargs) -> bool:
        self._check_key(key)
        config = self._get_scope_config(scope)
        key = config.normalize_uniq_column_values(key)
        return self._set(key, value, config)

    def pop(self, key: _KEY_TYPE, scope: str = None, **kwargs) -> None:
        self._check_key(key)
        config = self._get_scope_config(scope)
        key = config.normalize_uniq_column_values(key)
        self._pop(key, config)

    def scopes(self) -> Iterable[str]:
        return self._scopes.keys()

    def keys(self, scope: str = None) -> Iterable[str]:
        config = self._get_scope_config(scope)
        operation = self.sql_builder.build_select_operation(config=config, key=None)
        return self._execute(operation, fetch_size='all', cursor='new')

class SingleTableCache(BaseRelationalDbCache):
    """
    Use single table in the db as backend.
    This is useful when data of different scopes stored in the same table.
    """

    @classmethod
    def from_uri(cls, uri: str, **kwargs) -> 'SingleTableCache':
        assert re.match(r'\w+://.+', uri), uri
        schema, database = uri.split('://')
        return cls(database=database, **kwargs)

    def __init__(self, connection=None, database: str = None, connect_args: Dict[str, Any] = None,
                 config: SingleTableConfig = None):
        """
        Create a cache based on relational database.
        :param connection:          connection to the database
        :param database:            uri for the database
        :param connect_args:        custom connect args
        :param config:              config columns for scope and uniq columns
        """
        super().__init__(connection, database, connect_args)
        self.config = config
        self._create_table_if_not_exists(config)

    def _check_key(self, key: _KEY_TYPE, scope: _KEY_TYPE):
        if key is None:
            raise ValueError("Arg key should not be None.")
        if scope is None:
            raise ValueError("Arg scope should not be None.")

    def _gen_uniq_column_values(self, key: _KEY_TYPE, scope: _KEY_TYPE) -> Sequence[Any]:
        self._check_key(key, scope)
        key = self.config.normalize_key_column_values(key)
        scope = self.config.normalize_scope_column_values(scope)
        return tuple(scope) + tuple(key)

    def exists(self, key: _KEY_TYPE, scope: _KEY_TYPE = None, **kwargs) -> bool:
        uniq_column_values = self._gen_uniq_column_values(key, scope)
        return self._exists(uniq_column_values, self.config)

    def fetch(self, key: _KEY_TYPE, default: Any = None, scope: _KEY_TYPE = None, **kwargs) -> Any:
        uniq_column_values = self._gen_uniq_column_values(key, scope)
        return self._fetch(uniq_column_values, self.config, default=default)

    def set(self, key: _KEY_TYPE, value: Any, scope: _KEY_TYPE = None, **kwargs) -> Any:
        uniq_column_values = self._gen_uniq_column_values(key, scope)
        return self._set(uniq_column_values, value, self.config)

    def pop(self, key: _KEY_TYPE, scope: _KEY_TYPE = None, **kwargs) -> Any:
        uniq_column_values = self._gen_uniq_column_values(key, scope)
        return self._pop(uniq_column_values, self.config)

    def scopes(self) -> Iterable[Any]:
        config = self.config
        columns = config.scope_columns
        operation = self.sql_builder.build_select_operation(config, select_columns=columns, distinct=True)
        row = self._execute(operation, fetch_size='all')
        if row is None:
            return []
        if len(columns) == 1:
            return [x[0] for x in row]
        return [tuple(x) for x in row]

    def keys(self, scope: _KEY_TYPE = None) -> Iterable[Any]:
        scope = self.config.normalize_scope_column_values(scope)
        config = self.config
        select_columns = config.key_columns
        condition_columns = config.scope_columns
        operation = self.sql_builder.build_select_operation(
            config, key=scope,
            select_columns=select_columns,
            condition_columns=condition_columns
        )
        row = self._execute(operation, fetch_size='all')
        if row is None:
            return []
        if len(select_columns) == 1:
            return [x[0] for x in row]
        return [tuple(x) for x in row]




# -*- coding: utf-8 -*-

from typing import (
    Dict, Any, Tuple, Sequence,
    Hashable, Union, Literal,
)
import warnings
from abc import ABC, abstractmethod
from .relational import (
    FormatSqlBuilder,
    BaseRelationalDbCache,
    RelationalDbScopeConfig,
    RelationalDbCache,
    SingleTableCache,
    Operation, ValueAdapter,
)

class MysqlSqlBuilder(FormatSqlBuilder):

    default_column_type = 'text'

    def get_value_type(self, value: Any) -> str:
        return 's'

    def build_select_all_table_operation(self) -> Operation:
        return Operation(statement="SHOW TABLES")

    def build_select_table_columns_operation(self, table: str, filter_uniq=False) -> Operation:
        if filter_uniq:
            stmt = f"SHOW INDEX FROM {table} WHERE key_name = 'PRIMARY'"
        else:
            stmt = f"DESCRIBE {table}"
        return Operation(statement=stmt)

    def get_column_name_from_result(self, result: Sequence[Any]) -> str:
        # 4th is column name.
        return result[4]

    def _gen_update_statement(self, config: 'MysqlScopeConfig', value: Dict[str, Any]) -> Tuple[str, bool]:
        first = ','.join(self.config_variable(name=k, order=i, value=v)
                         for i, (k, v) in enumerate(value.items(), 1))
        second = ','.join(f'{k}={self.config_variable(name=k, order=i, value=v)}'
                          for i, (k, v) in enumerate(value.items(), 1))
        return (f"INSERT INTO {config.table}({','.join(value.keys())})"
                f" VALUES ({first})"
                f" ON DUPLICATE KEY"
                f" UPDATE {second}"), False

class MysqlScopeConfig(RelationalDbScopeConfig):
    def __init__(self, scope: Hashable, table: str = None, uniq_column: str = None,
                 uniq_columns: Sequence[str] = ('id',), columns: Sequence[str] = ('data',),
                 column_types: Dict[str, str] = None,
                 value_adapter: Union[Literal['auto', 'default', 'single', 'tuple', 'list'], ValueAdapter] = 'auto',
                 default_column_type: str = None):
        warnings.warn("use RelationalDbScopeConfig instead", DeprecationWarning)
        super().__init__(scope, table, uniq_column, uniq_columns, columns, column_types, value_adapter,
                         default_column_type)

class BaseBackend(BaseRelationalDbCache, ABC):
    """
    Base abstract class for mysql backend.
    """
    paramstyle = 'format'
    sql_builder = MysqlSqlBuilder()

    @abstractmethod
    def _connect_db0(self, **kwargs):
        """
        The actual method to connect database .
        """
        pass

    def connect_db(self, database: str = None, connect_args: Dict[str, Any] = None):
        kwargs = dict(connect_args or {})
        if database is not None:
            # TODO 2023/10/12  parse uri
            import warnings
            warnings.warn("uri is ignored")
            return self._connect_db0(**kwargs)

class PymysqlBackend(BaseBackend, ABC):
    def _connect_db0(self, **kwargs):
        import pymysql
        return pymysql.connect(**kwargs)

class MysqlConnectorBackend(BaseBackend, ABC):
    def _connect_db0(self, **kwargs):
        import mysql.connector
        return mysql.connector.connect(**kwargs)

try:
    import pymysql
    BaseMysqlCache = PymysqlBackend
except ImportError:
    import mysql.connector
    BaseMysqlCache = MysqlConnectorBackend

class MysqlCache(BaseMysqlCache, RelationalDbCache):
    """
    Use a mysql database as backend.
    """

class MysqlSingleTableCache(BaseMysqlCache, SingleTableCache):
    """
    Use one table of mysql as backend.
    """
